import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

import torchvision
import torchvision.transforms as transforms

from torch.utils.tensorboard import SummaryWriter

from collections import OrderedDict
from collections import namedtuple
from itertools import product  

torch.set_printoptions(linewidth=120)  
torch.set_grad_enabled(True)  
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu") 


params = OrderedDict(
    lr=[0.0001, 0.001, 0.01],
    batch_size=[10, 100, 1000, 10000],
    shuffle=[False, True]
)


def get_num_correct(preds, labels):
    return preds.argmax(dim=1).eq(labels).sum().item()


@torch.no_grad()
def get_all_preds(model, loader):
    all_preds = torch.tensor([])
    for batch in loader:
        images, labels = batch
        preds = model(images).cpu()

        all_preds = torch.cat(
            (all_preds, preds),
            dim=0
        )
    return all_preds


class RunBuilder:

    @staticmethod
    def get_runs(params):
        run = namedtuple('run', params.keys())
        runs = []
        for v in product(*params.values()):
            runs.append(run(*v))
        return runs


class Network(nn.Module):
    def __init__(self):
        super(Network, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=1, out_channels=6, kernel_size=5)
        self.conv2 = nn.Conv2d(in_channels=6, out_channels=12, kernel_size=5)

        self.fc1 = nn.Linear(in_features=12*4*4, out_features=120)
        self.fc2 = nn.Linear(in_features=120, out_features=60)
        self.out = nn.Linear(in_features=60, out_features=10)

    def forward(self, t):
        t = F.max_pool2d(F.relu(self.conv1(t)), kernel_size=2, stride=2) 
        t = F.max_pool2d(F.relu(self.conv2(t)), kernel_size=2, stride=2) 
        t = F.relu(self.fc1(t.reshape(-1, 192)))
        t = F.relu(self.fc2(t))
        t = self.out(t)
        return t


def main():
    print('Torch version: ', torch.__version__)

    train_set = torchvision.datasets.FashionMNIST(
        root='./data/',  
        train=True,  
        download=True,  
        transform=transforms.Compose([  
            transforms.ToTensor()  
        ])
    )

    for run in RunBuilder.get_runs(params):
        print(f'-{run}')

        train_loader = torch.utils.data.DataLoader(
            train_set, 
            batch_size=run.batch_size,
            shuffle=run.shuffle
        )

        network = Network()
        if torch.cuda.is_available():
            network = nn.DataParallel(network)
        optimizer = optim.Adam(network.parameters(), lr=run.lr)

        images, labels = next(iter(train_loader))
        grid = torchvision.utils.make_grid(images)

        comment = f'-{run}' 
        tb = SummaryWriter(comment=comment)
        tb.add_image('Images', grid)
        tb.add_graph(network, images) # graph one batch per run

        for epoch in range(1):
            total_loss = 0
            total_correct = 0
            for batch in train_loader:
                if torch.cuda.is_available():
                    images, labels = batch[0].to(device), batch[1].to(device)
                else:
                    images, labels = batch

                preds = network(images)
                loss = F.cross_entropy(preds, labels)
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                total_loss += loss.item() * run.batch_size  # used to compare losses of different batch size runs
                total_correct += get_num_correct(preds, labels)

            tb.add_scalar('Loss', total_loss, epoch)
            tb.add_scalar('Total correct', total_correct, epoch)
            tb.add_scalar('Accuracy', total_correct/len(train_set), epoch)
            for name, weight in network.named_parameters():
                tb.add_histogram(name, weight, epoch)
                tb.add_histogram(f'{name}.grad', weight, epoch)

            print(
                "epoch:", epoch, "total correct:", total_correct,
                "loss:", total_loss, "accuracy: ", total_correct/len(train_set)
            )

    return


if __name__ == '__main__':
    main()
